###########################################################################
# Copyright (c) 2016 Franco (nextime) Lanza <franco@nexlab.it>
#
# Domotika System Controller Daemon "domotikad" 
#
# This file is part of domotikad.
#
# domotikad is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from zope.interface import Interface, implements
from twisted.python import components

from nevow import rend, inevow
from corepost import Response, NotFoundException, AlreadyExistsException
from corepost.web import RESTResource, route, Http
from corepost.convert import convertToJson
from corepost.enums import MediaType, HttpHeader

from nexlibs.utils import webutils as wu

import hashlib
import hmac
import six
import time
from Queue import Queue
import uuid

from domotika.singleton import messengerlinks 
from StringIO import StringIO
import imghdr

_MESSENGER_LINKS = messengerlinks.MessengerLinkRegistry()
_MESSENGER_PSID = messengerlinks.MessengerPSIDRegistry()

try:
    #python2
    from urllib import urlencode
except ImportError:
    #python3
    from urllib.parse import urlencode

import logging
log = logging.getLogger( 'Webgui' )


BOTResource = RESTResource



class IBotProtoInterface(Interface):
 
   def getUsername(self, uid):
      """ get a username from uid """
 
   def isLogged(self, uid):
      """ return True if it's a know user """
 
   def sendAPI(self, payload, req_uri):
      """ Send request to the API """

   def receivedAuthentication(self, msg):
      """ receive auth message """

   def receivedTextMessage(self, msg):
      """ receive text message """

   def receivedDeliveryConfirmation(self, msg):
      """ receive delivery confirmation """

   def receivedPostback(self, msg):
      """ receive a postback message """

   def receivedAccountLink(self, msg):
      """ receive account linking message """

   def sendCommandList(self, recipient_id):
      """ send command list to the user """

   def sendScreenshotList(self, recipient_id):
      """ send a list of screenshots available """

   def sendScreenshot(self, recipient_id, target):
      """ send one or more screenshots """

   def sendClima(self, recipient_id):
      """ send Clima status """

   def sendMessage(self, recipient_id, message):
      """ send a text message """

   def sendImageMessage(self, recipient_id, imguri):
      """ send an image link """

   def sendImageMessageData(self, recipient_id, imagedata, mtype):
      """ send an image by data """

   def sendAuthRequest(self, recipient_id):
      """ send login request """



class BotProto(object):

   def __init__(self, core):
      self.core = core



def botCommandParser(self, uid, txt):
   def voiceResult(res):
      log.info('VoiceResult: '+str(res))
      if len(res) > 0:
         if res[0] == 'Ok' and len(res) > 1:
            try:
               result = 'ho eseguito "'+" ".join(res[1]['clean'])+'"'
            except:
               pass
         else:
            result = 'Spiacente, non ho trovato un comando corrispondente'
         self.sendMessage(uid, result)

   txt = unicode(txt.lower())
   if not self.isLogged(uid):
      if txt == u'hello' or txt == u'ciao':
         self.sendMessage(uid, 'Hello, how can i help you?')
      elif txt == u'?' or txt == u'help':
         self.sendMessage(uid, 'you can\'t do that.')
      elif txt == u'login':
         self.sendAuthRequest(uid)
      else:
         self.sendMessage(uid, txt+u" come se fosse antani")

   else:
      if txt == u'hello' or txt == u'ciao':
         self.sendMessage(uid, 'Hello %s, how can i help you?' %(self.getUsername(uid)))
      elif txt == u'?' or txt == u'help':
         self.sendMessage(uid, 'Ok, devo ancora implementare l\'aiuto! Sorry for that!')
         self.sendMessage(uid, 'Anyway, i comandi che hai sono: "command list", "screenshot list", "screenshot" e "clima".')
         self.sendMessage(uid, 'Qualsiasi altro comando viene interpretato come un potenziale comando vocale.')
      elif txt == u'logout':
         self.sendMessage(uid, 'Ok, devo ancora implementare anche il logout')
      elif txt == u'login':
         self.sendMessage(uid, 'Sei gia\' loggato, %s!' %(self.getUsername(uid)))
      elif txt == u'command list' or txt == 'cmd list':
         self.sendCommandList(uid)
      elif txt == u'screenshot list' or txt==u'ss list':
         self.sendScreenshotList(uid)
      elif txt.startswith('screenshot ') or txt.startswith('ss '):
         self.sendScreenshot(uid, " ".join(txt.split()[1:]))
      elif txt == u'clima':
         self.sendClima(uid)
      else:
         self.core.voiceReceived(txt, confidence=1.0, lang="it").addCallback(voiceResult)


class BotCore(object):
   path = ""
   
   def __init__(self, core, session):
      self.core = core
      self.session = session


class BaseBot(BotCore):

   @route("/")
   def welcome(self, request, *a, **kw):
      return 'Domotika BOTs API interface V1.0'


def messengerValidator():
   def checkRequest(f):
      def new_f(self, request, *a, **kw):
         if 'x-hub-signature' in request.getAllHeaders().keys():
            hubsig=request.getHeader('x-hub-signature')  
         try:
            hash_method, hub_signature = hubsig.split('=')
         except:
            pass
         else:
            digest_module = getattr(hashlib, hash_method)
            hmac_object = hmac.new(str(self._cfgGet('app_secret')), unicode(request.content.getvalue()), digest_module)
            generated_hash = hmac_object.hexdigest()
            if hub_signature == generated_hash:
               return f(self, request, *a, **kw)
         return Response(200, "Failed validation." )
      new_f.funct_name = f.func_name
      return new_f
   return checkRequest


class MessengerMessage(object):
   
   def __init__(self, headers, payload, uri):
      self.headers = headers
      self.payload = payload
      self.uri = uri

class MessengerConf(object):

   def _cfgGet(self, keyword):
      return self.core.configGet('messenger', keyword)


class MessengerBotAdapter(MessengerConf):

   implements(IBotProtoInterface)

   graphuri = 'https://graph.facebook.com/v2.6'
   sendQueue = Queue()
   sendlock = False

   def __init__(self, orig):
      self.orig = orig
      self.core = orig.core

   @property
   def auth_args(self):
      if not hasattr(self, '_auth_args'):
         auth = {
            'access_token': unicode(self._cfgGet('page_token'))
         }
         if self._cfgGet('app_secret') is not None and len(self._cfgGet('app_secret'))>0:
            appsecret_proof = self._generateAppSecProof()
            auth['appsecret_proof'] = appsecret_proof
         self._auth_args = auth
      return self._auth_args

   def _generateAppSecProof(self):
      if six.PY2:
         hmac_object = hmac.new(str(self._cfgGet('app_secret')), unicode(self._cfgGet('page_token')), hashlib.sha256)
      else:
         hmac_object = hmac.new(bytearray(self._cfgGet('app_secret'), 'utf8'), str(self._cfgGet('page_token')).encode('utf8'), hashlib.sha256)
      return hmac_object.hexdigest()


   def _dataSent(self, res):
      log.debug('Messenger BOT Datasent OK')
      log.debug(res)

   def _dataError(self, res):
      log.info('Messenger BOT Datasent ERROR')
      log.error(res)
      try:
         log.debug(res.value.reasons[0].printTraceback())
      except:
         pass


   def _sendRaw(self):
      if not self.sendQueue.empty() and not self.sendLock:
         message = self.sendQueue.get()
         return wu.getPage(message.uri, method='POST', headers=message.headers,
            postdata=message.payload).addCallbacks(self._dataSent, self._dataError)

   def sendAPI(self, payload, req_uri='/me/messages'):
      request_endpoint = self.graphuri+req_uri
      request_uri = request_endpoint+'?'+urlencode(self.auth_args)
      log.info("Sending message to "+str(request_uri)+" : "+str(convertToJson(payload)))
      ctype={"Content-Type": "application/json"}
      return wu.getPage(request_uri, method='POST', headers=ctype, postdata=convertToJson(payload)).addCallbacks(self._dataSent, self._dataError)

   def getUsername(self, uid):
      return _MESSENGER_PSID.get_link(uid)     

   def isLogged(self, uid):
      return  _MESSENGER_PSID.linkid_exists(uid)

   def receivedAuthentication(self, msg):
      log.info("Messenger bot received authentication: "+str(msg))

   def receivedTextMessage(self, msg):

      log.info("Messenger bot received message: "+str(msg))
      senderid = msg['sender']['id']
      if not 'is_echo' in msg['message'].keys() or not msg['message']['is_echo']:
         message = u'Received: '+unicode(msg['message']['text'])
         txt = unicode(msg['message']['text']).lower()
         botCommandParser(self, senderid, txt)

   def receivedDeliveryConfirmation(self, msg):
      log.info("Messenger bot received delivery confirmation: "+str(msg))

   def receivedPostback(self, msg):
      log.info("Messenger bot received postback: "+str(msg))

   def receivedMessageRead(self, msg):
      log.info("Messenger bot received message read: "+str(msg))

   def receivedAccountLink(self, msg):
      log.info("Messenger bot received accountlink: "+str(msg))
      senderid = msg['sender']['id']
      recipient = msg['recipient']['id']
      status = msg['account_linking']['status']
      authcode = msg['account_linking']['authorization_code']
      if status == u'linked':
         self.core.add_messenger_psid(senderid, authcode)

   def sendCommandList(self, senderid):
      def pushList(res):
         for r in res:
            self.sendMessage(senderid, " * "+str(r[0]))
      self.sendMessage(senderid, "Target list for string/voice commands (call them with the right action like up,down,open,close and so on.):")
      self.core.getVoiceCommandList().addCallback(pushList)

   def sendScreenshotList(self, recipient_id):
      def pushList(res):
         for r in res:
            self.sendMessage(recipient_id, " * "+str(r[0]))
      self.sendMessage(recipient_id, "Cam list:")
      self.core.getScreenshotList().addCallback(pushList)


   def sendScreenshot(self, recipient_id, target):
      def pushImage(res):
         if res:
            self.sendImageMessageData(recipient_id, res)
         else:
            self.sendMessage(recipient_id, 'Cannot retrieve image of '+str(target))
      if target.endswith('%'):
         target=target[:-1]
      if len(target) > 0:
         self.core.getScreenshot(target, pushImage)

   def sendClima(self, recipient_id):
      def pushClima(res):
         for r in res:
            if r[0] == 'last.rain':
               msg=r[0]+": "+time.strftime('%d-%m-%Y %H:%M:%S', time.localtime(int(r[1])))
            else:
               msg=r[0]+": "+str(r[1])
            self.sendMessage(recipient_id, msg)
      def pushThermo(res):
         for r in res:
            msg=r[0]+": set "+str(r[1])+" read "+str(float(r[2])/10)+" function "+r[3]
            self.sendMessage(recipient_id, msg)
         return self.core.getClimaUniques().addCallback(pushClima)
      self.core.getThermostats().addCallback(pushThermo)

   def sendMessage(self, recipient_id, message):
      payload = {
         'recipient': {
            'id': recipient_id
         },
         'message': {
            'text': message
         }
      }
      return self.sendAPI(payload)

   def sendImageMessage(self, recipient_id, imageuri):
      payload = {
         'recipient': {
            'id': recipient_id
         },
         'message': {
            'attachment': {
               'type': 'image',
               'payload': {
                  'url': imageuri
               }
            }
         }
      }
      return self.sendAPI(payload)

   def sendImageMessageData(self, recipient_id, imagedata, mtype=None):

      if not imagedata:
         return
      if mtype == None:
         ft = imghdr.what(None, imagedata)
         log.info("FILE TYPE: "+str(ft))
         if ft in ['jpeg', 'gif', 'png']:
            mtype=ft
         else:
            log.warning("Image mimetype doesn't seems to be valid: "+str(ft))
            log.warning("Assuming it's a jpeg.")
            mtype='jpeg'


      fname=str(uuid.uuid4().get_hex())+mtype
      bond="------------------------"+str(uuid.uuid4().get_hex())[:16]
      data="--"+bond+"\r\n"
      data+="Content-Disposition: form-data; name=\"recipient\"\r\n\r\n"
      data+="{\"id\": \""+str(recipient_id)+"\"}\r\n"
      data+="--"+bond+"\r\n"
      data+="Content-Disposition: form-data; name=\"message\"\r\n\r\n"
      data+='{"attachment":{"type":"image", "payload":{}}}'+"\r\n"
      data+="--"+bond+"\r\n"
      data+='Content-Disposition: form-data; name="filedata"; filename="'+fname+'"'+"\r\n"
      data+='Content-Type: image/'+mtype+"\r\n\r\n"
      data+=str(imagedata)
      data+="\r\n--"+bond+"--\r\n"
      headers={
         'Accept': '*/*',
         'Content-Type': 'multipart/form-data; boundary='+bond
      }

      request_endpoint=self.graphuri+'/me/messages'
      request_uri = request_endpoint+'?'+urlencode(self.auth_args)
      return wu.getPage(request_uri, agent="curl/7.50.1",  method='POST', headers=headers,
         postdata=data, expect100=True).addCallbacks(self._dataSent, self._dataError)

   def sendAuthRequest(self, recipient_id):
      payload = {
         'recipient': {
            'id': recipient_id
         },
         'message': {
            "attachment": {
               "type": "template",
               "payload": {
                  "template_type": "generic",
                  "elements": [{
                     "title": self._cfgGet('auth_title'),
                     "image_url": self._cfgGet('auth_img'),
                     "buttons": [{
                        "type": "account_link",
                        "url": self._cfgGet('auth_host')
                     }]
                  }]
               }
            }
         }
      }
      return self.sendAPI(payload)
   
components.registerAdapter(
   MessengerBotAdapter,
   BotProto,
   IBotProtoInterface
)

class MessengerBot(BotCore, MessengerConf):

   path = "messenger"

   @route("/", Http.GET)
   def rootGet(self, request, *a, **kw):
      if 'hub.mode' in kw.keys() and kw['hub.mode'] == 'subscribe' and 'hub.challenge' in kw.keys():
         if 'hub.verify_token' in kw.keys() and kw['hub.verify_token'] == self._cfgGet('verify_token'):
            log.info("New verification request for Messenger BOT: "+str(kw))
            return kw['hub.challenge']
      return Response(403, "Failed validation." )

   @route("/", Http.POST)
   @messengerValidator()
   def rootPost(self, request, *a, **kw):
      log.debug("New messenger bot request: "+str(kw))
      botproto = BotProto(self.core)
      botiface = IBotProtoInterface(botproto)
      if 'object' in kw.keys() and kw['object'] == 'page' and 'entry' in kw.keys():
         for entry in kw['entry']:
            #try:
            if True:
               pageID=entry['id']
               timestamp=entry['time']
               messaging=entry['messaging']
               for message in messaging:
                  msgkeys = message.keys()
                  if 'optin' in msgkeys:
                     botiface.receivedAuthentication(message)
                  elif 'message' in msgkeys:
                     botiface.receivedTextMessage(message)
                  elif 'delivery' in msgkeys:
                     botiface.receivedDeliveryConfirmation(message)
                  elif 'postback' in msgkeys:
                     botiface.receivedPostback(message)
                  elif 'read' in msgkeys:
                     botiface.receivedMessageRead(message)
                  elif 'account_linking' in msgkeys:
                     botiface.receivedAccountLink(message)
                  else:
                     log.info("Received unknow messaging webhook for messenger BOT: "+str(kw))
            #except:
            #   pass

      return Response(200, 'OK')


   @route("/loginfailed")
   def loginFailed(self, request, *a, **kw):
      k = kw.keys()
      if 'account_linking_token' in k and 'redirect_uri' in k:
         log.info("Passed linking uri "+kw['redirect_uri'])
         luri=kw['redirect_uri']
         log.info("Authorization failed: redirecting to "+luri)
         return Response(302, 'Authorization failed', headers={'Location': luri})
      return Response(200, "Authorization failed." )
  


class MessengerBotAuth(BotCore, MessengerConf):

   path="messenger"

   @route("/", Http.GET)
   def linkUser(self, request, *a, **kw):
      k = self.session.saveargs.keys()
      if 'account_linking_token' in k and 'redirect_uri' in k:
         luri=self.session.saveargs['redirect_uri'][0]
         log.debug(str(self.session.mind.perms))
         hmac_object = hmac.new(str(self.session.mind.perms.username), unicode(self._cfgGet('page_token')), hashlib.sha256)
         _MESSENGER_LINKS.add_link(hmac_object.hexdigest(), self.session.mind.perms)
         luri = luri+"&authorization_code="+hmac_object.hexdigest()
         log.info("Authorization succeed: redirecting to "+luri)
         headers={'Location': luri}
         return Response(302, 'Authorization succeed', headers=headers)
      return Response(200, "Authorization Succeed." )



BotApiList=(
   BaseBot,
   MessengerBot
)


class BotPages(rend.Page):


   def child_(self, ctx):
      request = inevow.IRequest(ctx)
      self.session = inevow.ISession(ctx)
      request.setHeader("pragma", "no-cache")
      request.postpath=['/']
      return BOTResource((BaseBot(self.core, self.session),))

   def childFactory(self, ctx, name):
      request = inevow.IRequest(ctx)
      self.session = inevow.ISession(ctx)
      request.setHeader("pragma", "no-cache")
      #request.postpath=['/', name]+request.postpath
      request.postpath=['/']+request.postpath
      log.info("BOT: "+str(request.postpath))
      return BOTResource([x(self.core, self.session) for x in BotApiList])


BotAuthList=(
   BaseBot,
   MessengerBotAuth
)

class BotAuth(rend.Page):

   def childFactory(self, ctx, name):
      request = inevow.IRequest(ctx)
      self.session = inevow.ISession(ctx)
      request.setHeader("pragma", "no-cache")
      #request.postpath=['/', name]+request.postpath
      request.postpath=['/']+request.postpath
      log.info("BOTAUTH: "+str(request.postpath))
      return BOTResource([x(self.core, self.session) for x in BotAuthList])

